PROGRAM Metro (infile, OUTPUT);
  (* version 2.7 *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 100;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxchr = 200;     (* max. number characterstates per species *)
   bits = 30;        (* max. number of elements in a Pascal set *)
   wrds = 7;         (* wrds * bits >= maxchr                   *)
   nmlngth = 10;     (* number of characters in species name    *)
   stoptemp = 10.0;  (* final value of 1/kT                     *)
   howoften = 1;     (* when new tree found print line when     *)

TYPE steparray = ARRAY[1..maxchr] OF INTEGER;
   bitrange = 1..bits;
   bitset = SET OF bitrange;
   bitarray = ARRAY[1..wrds] OF bitset;
   pointer = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : pointer;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      stateone, statezero : bitarray; (* see in PROCEDURE fillin *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF pointer;
   longer = ARRAY[0..5] OF INTEGER;

VAR root : pointer;
   infile : TEXT;
   spp, nonodes, chars, words, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of characteristics
        words = number of words needed
                to represent characteristics of one organism
        outgrno indicates outgroup *)
   weights, thresh, ancvar, questions,
   allsokal, allwagner, mixture, noroot, outgropt, didreroot : BOOLEAN;
   extras, weight : steparray;
   ancone, anczero : ARRAY[1..maxchr] OF BOOLEAN;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   delta, threshold : REAL;
   wagner : bitarray;
   seed : longer;

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
   BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)

   FUNCTION random (VAR seed : longer) : REAL;
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* random -- slow but machine independent *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO 
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;    
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      random := x;
   END; (* random *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
         VAR i, inseed : INTEGER;
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Mixed parsimony by annealing, version 2.7');
         WRITELN(OUTPUT);
         READ(infile, spp, chars, delta, inseed);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  characters');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Annealing parameter increment = ', 
                            delta:15:8);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Random number seed = ', inseed:8);
         WRITELN(OUTPUT);
         words := (chars DIV bits) + 1;
         nonodes := 2 * spp - 1;
         FOR i := 0 TO 5 DO seed[i] := 0;
         i := 0;
         REPEAT
            seed[i] := inseed MOD 64;
            inseed := inseed DIV 64;
            i := i + 1;
         UNTIL inseed = 0;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputmixture;
            (* input mixture of methods *)
            VAR i, j, k : INTEGER;
               ch : CHAR;
               wag : BOOLEAN;
         BEGIN (* inputmixture *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO words DO wagner[i] := [];
            j := 0;
            k := 1;
            FOR i := 1 TO chars DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
               uppercase (ch);
               wag := FALSE;
               IF (ch='W') OR (ch = '?')
               THEN wag := TRUE
               ELSE IF (ch='S') OR (ch='C')
                  THEN wag := FALSE
                  ELSE WRITELN('error: ', ch);
               j := j + 1;
               IF j > bits
               THEN BEGIN
                  j := 1;
                  k := k + 1;
                  END;
               IF wag
               THEN wagner[k] := wagner[k] + [j];
               IF NOT ancvar
               THEN ancone[i] := wag;
               END;
            READLN(infile);
            allwagner := FALSE;
            allsokal := FALSE;
            mixture := TRUE;
         END; (* inputmixture *)

         PROCEDURE printmixture;
            (* print out list of parsimony methods *)
            VAR i, k, l : INTEGER;
         BEGIN (* printmixture *)
            allwagner := TRUE;
            WRITELN(OUTPUT, ' Parsimony methods:');
            l := 0;
            k := 1;
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  k := k + 1;
                  END;
               IF l IN wagner[k]
               THEN WRITE(OUTPUT, 'W')
               ELSE BEGIN
                  WRITE(OUTPUT, 'S');
                  allwagner := allwagner AND (weight[i] = 0);
                  END;
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printmixture *)

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(OUTPUT, ' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of characters *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'    Characters are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

         PROCEDURE inputancestors;
            (* reads the ancestral states for each character *)
            VAR i : INTEGER;
               ch : CHAR;
         BEGIN (* inputancestors *)
            ancvar := TRUE; 
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
               ancone[i] := TRUE;
               anczero[i] := TRUE;
               IF ch = 'p' THEN ch := 'P';
               IF ch = 'b' THEN ch := 'B';
               CASE ch OF
                  '1' : anczero[i] := FALSE;
                  '0' : ancone[i] := FALSE;
                  'P' : BEGIN END;
                  'B' : BEGIN END;
                  '?' : BEGIN END
                  END;
               END;
            READLN(infile);
         END; (* inputancestors *)

         PROCEDURE printancestors;
            (* print out list of ancestral states *)
            VAR i : INTEGER;
         BEGIN (* printancestors *)
            WRITELN(OUTPUT, ' Ancestral states:');
            FOR i := 1 TO nmlngth+4 DO WRITE(OUTPUT, ' ');
            FOR i := 1 TO chars DO BEGIN
               newline (i, 50, nmlngth+4);
               IF ancone[i] AND anczero[i]
               THEN WRITE(OUTPUT, '?')
               ELSE IF ancone[i]
                  THEN WRITE(OUTPUT, '1')
                  ELSE WRITE(OUTPUT, '0');
               IF ((i MOD 5) = 0) THEN WRITE(OUTPUT, ' ');
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
         END; (* printancestor *)

      BEGIN (* inputoptions *)
         ancvar := FALSE;
         allwagner := TRUE;
         allsokal := FALSE;
         mixture := FALSE;
         FOR i := 1 TO chars DO BEGIN
            ancone[i] := TRUE;
            anczero[i] := TRUE;
            END;
         weights := FALSE;
         thresh := FALSE;
         outgropt := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'A') OR (ch = 'B') OR (ch = 'C') OR (ch = 'M')
               OR (ch = 'O') OR (ch = 'S') OR (ch = 'T')
               OR (ch = 'W')
            THEN CASE ch OF
                  'A' : extranum := extranum + 1;
                  'C' : allsokal := TRUE;
                  'M' : extranum := extranum + 1;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'S' : allsokal := TRUE;
                  'T' : extranum := extranum + 1;
                  'W' : extranum := extranum + 1
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         IF allsokal THEN allwagner := FALSE;
         FOR i := 1 TO words DO
            IF allsokal
            THEN wagner[i] := []
            ELSE wagner[i] := [1..bits];
         READLN(infile);
         threshold := 35 * spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'A'
            THEN inputancestors;
            IF ch = 'M'
            THEN inputmixture;
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'W'
            THEN inputweights;
            IF ch = 'T'
            THEN inputhresh;
            END;
         IF allsokal
         THEN BEGIN
            IF NOT ancvar
            THEN FOR i := 1 TO chars DO ancone[i] := FALSE;
            WRITELN(OUTPUT, ' Camin-Sokal parsimony method');
            END;
         IF allwagner
         THEN WRITELN(OUTPUT, ' Wagner parsimony method');
         WRITELN(OUTPUT);
         IF mixture
         THEN printmixture;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights;
         IF ancvar
         THEN printancestors;
         noroot := allwagner OR (threshold <= 2.0);
         questions := FALSE;
         FOR i := 1 TO chars DO BEGIN
            noroot := noroot AND ancone[i] AND anczero[i];
            questions := questions OR (ancone[i] AND anczero[i]);
            END;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and character state data for species *)
         VAR i, j, l : INTEGER;
            k : bitrange;
            charstate : CHAR;  (* possible states are
                                  '0', '1', 'P', 'B', and '?' *)
      BEGIN (* inputdata *)
         WRITELN(OUTPUT, ' Character-state data:');
         WRITELN(OUTPUT);
         FOR i := 1 TO chars DO extras[i] := 0;
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO words DO BEGIN
                  treenode[i]^.stateone[j] := [];
                  treenode[i]^.statezero[j] := [];
                  END;
               FOR j := 1 TO chars DO BEGIN
                  k := ((j-1) MOD bits) + 1;
                  l := ((j-1) DIV bits) + 1;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, charstate);
                  IF charstate = 'b' THEN charstate := 'B';
                  IF charstate = 'p' THEN charstate := 'P';
                  IF NOT ((charstate = '0')
                         OR (charstate = '1')
                         OR (charstate = '?')
                         OR (charstate = 'P')
                         OR (charstate = 'B'))
                  THEN WRITELN(OUTPUT, ' Warning -- ',
                        'bad character state: ',charstate);
                  newline (j, 50, nmlngth+4);
                  WRITE(OUTPUT, charstate);
                  IF (j MOD 5) = 0 THEN WRITE(OUTPUT, ' ');
                  IF (charstate = '1')
                  THEN treenode[i]^.stateone[l]
                             := treenode[i]^.stateone[l] + [k];
                  IF (charstate = '0')
                  THEN treenode[i]^.statezero[l]
                             := treenode[i]^.statezero[l] + [k];
                  IF (charstate = 'P') OR (charstate = 'B')
                  THEN extras[j] := extras[j] + weight[j];
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         Then rearranges the tree for greatest "likelihood" *)
      VAR i : INTEGER;
         alpha, like, bestyet : REAL;
         guess, bestguess : ARRAY[1..maxchr] OF CHAR;
         there, item : pointer;
         numsteps, besteps : steparray;
         fullset : bitset;
         steps, zeroanc, oneanc : bitarray;

      PROCEDURE add (VAR below, newtip, newfork : pointer);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : pointer);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : pointer;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)
ee two statesets.
            stateone and statezero are the sets of character
            states that must be 1 or must be 0, respectively,
            in a most parsimonious reconstruction, based on the
            information at or above this node.  Note that this
            state assignment may change based on information further
            down the tree.  If a character is in both sets it is in
            state "P".  If in neither, it is "?". *)
         VAR i : INTEGER;
      BEGIN (* fillin *)
         FOR i := 1 TO words DO BEGIN
            steps[i] := (p^.leftdesc^.stateone[i]
                         * p^.rtdesc^.statezero[i])
                       + (p^.leftdesc^.statezero[i]
                          * p^.rtdesc^.stateone[i]);
            p^.stateone[i] := (p^.leftdesc^.stateone[i]
                + p^.rtdesc^.stateone[i])
               - (steps[i]*(wagner[i]+zeroanc[i]));
            p^.statezero[i] := (p^.leftdesc^.statezero[i]
                + p^.rtdesc^.statezero[i])
               - (steps[i]*(wagner[i]+(fullset-zeroanc[i])));
            END;
      END; (* fillin *)

      PROCEDURE evaluate (r: pointer);
         (* Determines the number of steps needed for a tree.
            This is the minimum number needed to evolve chars on
            this tree *)
         VAR i, stepnum, smaller : INTEGER;
            sum : REAL;
            numsone, numszero : steparray;

         PROCEDURE count (VAR stps : bitarray);
            (* counts the number of steps in a fork of the tree *)
            VAR i, j, l : INTEGER;
         BEGIN (* count *)
            j := 1;
            l := 0;
            FOR i := 1 TO chars DO BEGIN
               l := l + 1;
               IF l > bits
               THEN BEGIN
                  l := 1;
                  j := j + 1;
                  END;
               IF l IN stps[j]
               THEN BEGIN
                  IF l IN zeroanc[j]
                  THEN numszero[i] := numszero[i] + weight[i]
                  ELSE numsone[i] := numsone[i] + weight[i];
                  END;
               END;
         END; (* count *)

         PROCEDURE postorder (p : pointer);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)
         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               count (steps);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            numszero[i] := 0;
            numsone[i] := 0;
            END;
         FOR i := 1 TO words DO
            zeroanc[i] := fullset;
         postorder (r);
         count(r^.stateone);
         FOR i := 1 TO words DO
            zeroanc[i] := [];
         postorder(r);
         count(r^.statezero);
         FOR i := 1 TO chars DO BEGIN
            smaller := spp * weight[i];
            IF anczero[i]
            THEN BEGIN
               numsteps[i] := numszero[i];
               smaller := numszero[i];
               END;
            IF ancone[i] AND (numsone[i] < smaller)
            THEN numsteps[i] := numsone[i];
            stepnum := numsteps[i]+extras[i];
            IF stepnum <= threshold * weight[i]
            THEN sum := sum + stepnum
            ELSE sum := sum + threshold * weight[i];
            guess[i] := '?';
            IF (NOT ancone[i]) OR
               (anczero[i] AND (numszero[i] < numsone[i]))
            THEN guess[i] := '0'
            ELSE IF (NOT anczero[i]) OR
                  (ancone[i] AND (numsone[i] < numszero[i]))
               THEN guess[i] := '1';
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE rearrange (VAR r: pointer);
         (* traverses tree trying local rearrrangements *)
         VAR n : INTEGER;

         PROCEDURE tryrearr (p : pointer);
            (* Evaluates one rearrangement of the tree.
               Uses Metropolis stochastic criterion for
               deciding whether to accept a rearrangement. *)
            VAR frombelow, whereto, forknode : pointer;
               prob, oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  IF NOT noroot OR 
                     (p^.ancestor^.ancestor^.ancestor <> NIL)
                  THEN BEGIN
                     oldlike := like;
                     IF p^.ancestor^.leftdesc = p
                     THEN frombelow := p^.ancestor^.rtdesc
                     ELSE frombelow := p^.ancestor^.leftdesc;
                     whereto := p^.ancestor^.ancestor;
                     forknode := p^.ancestor;
                     remove (p, forknode);
                     add (whereto, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     evaluate (r);
                     alpha := alpha + delta;
                     prob := 1.0 / (1.0 + EXP(alpha * (oldlike - like)));
                     IF random(seed) > prob
                     THEN BEGIN
                        remove (p, forknode);
                        add (frombelow, p, forknode);
                        IF r^.ancestor <> NIL THEN r := r^.ancestor;
                        like := oldlike;
                        END
                     ELSE BEGIN
                        n := n + 1;
                        IF n = howoften
                        THEN BEGIN
                           n := 0;
                           WRITELN(OUTPUT, ' ', -like:10:3, alpha:10:4);
                           END;
                        besteps := numsteps;
                        bestguess := guess;
                        END;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : pointer);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         n := 0;
         WHILE alpha-stoptemp < 0.0 DO 
            repreorder (r);
      END; (* rearrange *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : pointer);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : pointer;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p :=root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.index - spp >= 10
                     THEN WRITE(OUTPUT, (p^.index - spp):2)
                     ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (though rerooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END
         ELSE BEGIN
            WRITELN(OUTPUT, '  /');
            WRITELN(OUTPUT, ' /');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints number of steps and table of numbers of steps in
            each character *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, j, k : INTEGER;
               unknown : BOOLEAN;
               dohyp : bitarray;

            PROCEDURE filltrav (r : pointer);
               (* traverse to fill in interior node states *)
            BEGIN (* filltrav *)
               IF NOT r^.tip
               THEN BEGIN
                  filltrav (r^.leftdesc);
                  filltrav (r^.rtdesc);
                  fillin (r);
                  END;
            END; (* filltrav *)

            PROCEDURE hyptrav (r : pointer);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  bottom, maybe, nonzero : BOOLEAN;
                  zerobelow, onebelow : bitarray;

               PROCEDURE hyprint;
                  (* print out states at node *)
                  VAR i, j, k  : INTEGER;
                     l : bitrange;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF (noroot AND (NOT didreroot))
                     THEN WRITE(OUTPUT, '        ')
                     ELSE WRITE(OUTPUT, ' root   ');
                     END
                  ELSE WRITE(OUTPUT,
                       (r^.ancestor^.index - spp):4, '    ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, name[r^.index, i])
                  ELSE WRITE(OUTPUT, (r^.index - spp):4, '      ');
                  IF bottom
                  THEN 