
      PROCEDURE addpreorder (p, item : pointer);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

         PROCEDURE tryadd (p: pointer);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR nufork : pointer;
         BEGIN (* tryadd *)
            nufork := treenode[spp+i-1];
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            remove (item, nufork);
            IF nufork = root THEN root := p;
            IF like > bestyet THEN BEGIN
               bestyet := like;
               there := p;
               besteps := numsteps;
               bestguess := guess;
               END;
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item);
            addpreorder (p^.rtdesc, item);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: pointer);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : pointer);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : pointer;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := like;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     like := oldlike;
                     END
                  ELSE BEGIN
                     success := TRUE;
                     besteps := numsteps;
                     bestguess := guess;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : pointer);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : pointer);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : pointer;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               IF EOLN(infile) THEN READLN(infile);
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               IF NOT EOLN(infile) THEN READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT ((ch >= '0') AND (ch <= '9'));
               p := treenode[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[spp + 1];
         nextnode := spp;
         root^.ancestor := NIL;
         addelement (root);
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : pointer);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : pointer;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p :=root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.index - spp >= 10
                     THEN WRITE(OUTPUT, (p^.index - spp):2)
                     ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         IF noroot
         THEN BEGIN
            WRITELN(OUTPUT);
            WRITE(OUTPUT, '   remember:');
            IF didreroot
            THEN WRITE(OUTPUT, ' (though rerooted by outgroup)');
            WRITELN(OUTPUT, ' this is an unrooted tree!');
            END
         ELSE BEGIN
            WRITELN(OUTPUT, '  /');
            WRITELN(OUTPUT, ' /');
            END;
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints number of steps and table of numbers of steps in
            each character *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i, j, k : INTEGER;
               unknown : BOOLEAN;
               dohyp : bitarray;

            PROCEDURE filltrav (r : pointer);
               (* traverse to fill in interior node states *)
            BEGIN (* filltrav *)
               IF NOT r^.tip
               THEN BEGIN
                  filltrav (r^.leftdesc);
                  filltrav (r^.rtdesc);
                  fillin (r);
                  END;
            END; (* filltrav *)

            PROCEDURE hyptrav (r : pointer);
               (*  compute, print out states at one interior node *)
               VAR i : INTEGER;
                  bottom, maybe, nonzero : BOOLEAN;
                  zerobelow, onebelow : bitarray;

               PROCEDURE hyprint;
                  (* print out states at node *)
                  VAR i, j, k  : INTEGER;
                     l : bitrange;
               BEGIN (* hyprint *)
                  IF bottom
                  THEN BEGIN
                     IF (noroot AND (NOT didreroot))
                     THEN WRITE(OUTPUT, '        ')
                     ELSE WRITE(OUTPUT, ' root   ');
                     END
                  ELSE WRITE(OUTPUT,
                       (r^.ancestor^.index - spp):4, '    ');
                  IF r^.tip
                  THEN FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, name[r^.index, i])
                  ELSE WRITE(OUTPUT, (r^.index - spp):4, '      ');
                  IF bottom
                  THEN WRITE(OUTPUT, '          ')
                  ELSE IF nonzero
                     THEN WRITE(OUTPUT, '   yes    ')
                     ELSE If unknown
                        THEN WRITE(OUTPUT, '    ?     ')
                        ELSE IF maybe
                           THEN WRITE(OUTPUT, '  maybe   ')
                           ELSE WRITE(OUTPUT, '   no     ');
                  FOR j := 1 TO chars DO BEGIN
                     newline(j, 40, nmlngth+18);
                     k := (j-1) DIV bits + 1;
                     l := (j-1) MOD bits + 1;
                     IF (NOT (l IN wagner[k])) AND (bestguess[j] = '?')
                     THEN WRITE(OUTPUT, '.')
                     ELSE BEGIN
                        IF l IN r^.statezero[k]
                        THEN WRITE(OUTPUT, '0')
                        ELSE IF l IN r^.stateone[k]
                           THEN WRITE(OUTPUT, '1')
                           ELSE WRITE(OUTPUT, '?');
                        END;
                     IF (j MOD 5) = 0
                     THEN WRITE(OUTPUT, ' ');
                     END;
                  WRITELN(OUTPUT);
               END; (* hyprint *)

            BEGIN (* hyptrav *)
               bottom := (r^.ancestor = NIL);
               maybe := FALSE;
               nonzero := FALSE;
               IF bottom
               THEN BEGIN
                  zerobelow := zeroanc;
                  onebelow := oneanc;
                  END
               ELSE BEGIN
                  zerobelow := r^.ancestor^.statezero;
                  onebelow := r^.ancestor^.stateone;
                  END;
               FOR i := 1 TO words DO BEGIN
                  IF NOT r^.tip
                  THEN BEGIN
                     r^.statezero[i] := wagner[i]
                      * (zerobelow[i] * r^.leftdesc^.statezero[i]
                         + zerobelow[i] * r^.rtdesc^.statezero[i]
                         + r^.leftdesc^.statezero[i]
                          * r^.rtdesc^.statezero[i])
                      + dohyp[i] * (fullset-wagner[i])
                         * r^.statezero[i];
                     r^.stateone[i] := wagner[i]
                      * (onebelow[i] * r^.leftdesc^.stateone[i]
                         + onebelow[i] * r^.rtdesc^.stateone[i]
                         + r^.leftdesc^.stateone[i]
                          * r^.rtdesc^.stateone[i])
                       + dohyp[i] * (fullset-wagner[i])
                          * r^.stateone[i];
                     END;
                  maybe := maybe OR (dohyp[i] *
                          (r^.stateone[i] + r^.statezero[i])
                              <> (zerobelow[i] + onebelow[i]));
                  nonzero := nonzero OR ((r^.stateone[i] * zerobelow[i]
                              + r^.statezero[i] * onebelow[i]) <> []);
                  END;
               IF (NOT bottom) OR (NOT noroot) OR didreroot
               THEN hyprint;
               IF NOT r^.tip
               THEN BEGIN
                  hyptrav (r^.leftdesc);
                  hyptrav (r^.rtdesc);
                  END;
            END; (* hyptrav *)

         BEGIN (* hypstates *)
            FOR i := 1 TO words DO BEGIN
               zeroanc[i] := [];
               oneanc[i] := [];
               END;
            unknown := FALSE;
            FOR i := 1 TO chars DO BEGIN
               j := (i-1) DIV bits + 1;
               k := (i-1) MOD bits + 1;
               IF (bestguess[i] = '0')
               THEN zeroanc[j] := zeroanc[j] + [k];
               IF (bestguess[i] = '1')
               THEN oneanc[j] := oneanc[j] + [k];
               unknown := unknown OR ((NOT (k IN wagner[j]))
                                     AND (bestguess[i] = '?'));
               END;
            FOR i := 1 TO words DO
               dohyp[i] := wagner[i] + zeroanc[i] + oneanc[i];
            filltrav (root);
            WRITELN(OUTPUT);
            WRITE(OUTPUT, ' From    To     Any Steps?    ');
            WRITELN(OUTPUT, 'State at upper node');
            WRITELN(OUTPUT, ' ');
            hyptrav (root);
         END; (* hypstates *)

      BEGIN (* describe *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' requires a total of ',-like:10:3);
         WRITELN(OUTPUT);
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each character:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '      *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:6);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE WRITE(OUTPUT, besteps[k]+extras[k]:4);
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         IF questions AND ((NOT noroot) OR didreroot)
         THEN BEGIN
            WRITELN(OUTPUT, ' best guesses of ancestral states:');
            WRITE(OUTPUT, '       ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:2);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT, '      *--------------------');
            FOR i := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, ' ');
               WRITE(OUTPUT, 10*i:5, '!');
               FOR j := 0 TO 9 DO
                  IF (10*i+j = 0) OR (10*i+j > chars)
                  THEN WRITE(OUTPUT, '  ')
                  ELSE WRITE(OUTPUT, ' ', bestguess[10*i+j]);
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
            END;
         hypstates;
         WRITELN(OUTPUT);
      END; (* describe *)

      PROCEDURE reroot (outgroup : pointer);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : pointer;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

   BEGIN (* maketree *)
      fullset := [1..bits];
      IF NOT usertree
      THEN BEGIN
         add (treenode[2], treenode[1], treenode[spp+1]);
         root := treenode[spp+1];
         FOR i := 3 TO spp DO BEGIN
            bestyet := -32000;
            there := root;
            item := treenode[i];
            addpreorder (root, item);
            add (there, treenode[i], treenode[spp+i-1]);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            like := bestyet;
            rearrange (root);
            IF (NOT brief) OR (i = spp)
            THEN BEGIN
               didreroot := outgropt AND noroot AND (i = spp);
               IF didreroot
               THEN reroot (treenode[outgrno]);
               printree;
               IF (i = spp)
               THEN BEGIN
                  evaluate (root);
                  IF didreroot
                  THEN bestguess := guess;
                  describe;
                  END;
               END;
            END;
         END
      ELSE BEGIN
         WRITELN(OUTPUT, ' User-defined trees:');
         WRITELN(OUTPUT, ' ');
         READLN(infile, numtrees);
         FOR i := 1 TO numtrees DO BEGIN
            treeread;
            didreroot := outgropt AND noroot;
            IF didreroot
            THEN reroot (treenode[outgrno]);
            evaluate (root);
            besteps := numsteps;
            bestguess := guess;
            printree;
            describe;
            END;
         END;
   END; (* maketree *)

BEGIN (* Mixed parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree *)
   RESET(infile);
   doinput;
   maketree;
  (*  close(infile);  -- may need this here *)
END. (* Mixed parsimony by uphill search *)

;
                  describe;
                  END;
               END;
            END;
         END
      ELSE BEGIN
         WRITELN(OUTPUT, ' User-defined trees:');
         WRITELN(OUTPUT, ' ');
         READLN(infile, numtrees);
         FOR i := 1 TO numtrees DO BEGIN
            treeread;
            didreroot := outgropt AND noroot;
            IF didreroot
            THEN reroot (treenode[outgrno]);
            evaluate (root);
            besteps := numsteps;
            bestguess := guess;
            printree;
            describe;
            END;
         END;
   END; (* maketree *)

BEGIN (* Mixed parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree 