PROGRAM Plotgram (treefile, parmfile, fontfile, plotfile, OUTPUT);
   (* PHYLIP version 2.7
      Version 1.1, May 1985.  Written by Chris Meacham, c/o Herbarium,
      Department of Botany, University of California, Berkeley,
      California 94720, U.S.A.  *)
   CONST maxnodes = 100;
      maxlabels = 100;
      maxnch = 25;
      fontsize = 2905;    (* Fontsize of font1: 1035; font2: 1885;  *)
      bignum = 1000000.0; (*  font3: 2905; font4: 2012; font5: 2902 *)
   TYPE fontname = (font1, font2, font3, font4, font5);
      modetype = (yes, no);
      diagramtype = (phylogram, phenogram, cladogram);
      verticaltype = (levels, lengths, top, bottom, distance);
      horizontaltype = (straightline, nearest, furthest);
      drawingtype = (rectangle, slant);
      sidetype = (rightside, leftside);
      posoptiontype = (atroot, right, left, above, below, center);
      penstatustype = (penup, pendown);
      plottertype = (calcomp, tektronix, hp7470, bauschlomb, other);
      fonttype = ARRAY [1..fontsize] OF INTEGER;
      plotstring = ARRAY [1..maxnch] OF CHAR;
      nodeptr = ^treenode;
      labelptr = ^labelrecord;
      treenode = RECORD
         ancestor, sibling, descendant, lastdesc,
            left, right : nodeptr;
         nodenum, fromtop, frombottom : INTEGER;
         level, x, y, xstem, ystem : REAL;
         nodedone, popped : BOOLEAN;
         labels : labelptr;
         END;
      labelrecord = RECORD
         nextlab : labelptr;
         lab : plotstring;
         tipnum, nch  : INTEGER;
         xlab, ylab : REAL;
         END;

   VAR i, nnodes, nlabels, ntips, maxdepth, xnow, ynow,
         oldxhigh, oldxlow, oldyhigh, oldylow, popval : INTEGER;
      pi, rotation, xposition, yposition,
         xunitspercm, yunitspercm, aspectratio,
         xsize, ysize, xrange, yrange, labelheight,
         labelrotation, treeheight, treewidth, spacing,
         gap : REAL;
      cchex : ARRAY [0..15] OF CHAR;
      parmfile, treefile, fontfile, plotfile : TEXT;
      labelspresent, fontloaded, parmsok : BOOLEAN;
      diagram : diagramtype;
      vertical : verticaltype;
      horizontal : horizontaltype;
      drawing : drawingtype;
      fonttoload : fontname;
      font : fonttype;
      sizemode, leveltips, penchange, widthmode, popmode : modetype;
      plotter : plottertype;
      xoption, yoption : posoptiontype;
      root, garbnode : nodeptr;
      garblabel : labelptr;
      nodelist : ARRAY [1..maxnodes] OF nodeptr;
      labellist : ARRAY [1..maxlabels] OF labelptr;

   PROCEDURE initparms;
      CONST nabs = 25;
      VAR ab1, ab2 : ARRAY [0..nabs] OF CHAR;

      PROCEDURE plotrparms(ch : CHAR);
      BEGIN (* plotrparms *)
         CASE ch OF
            'C': BEGIN
               plotter := calcomp;
               xunitspercm := 39.37;
               yunitspercm := 39.37;
               xsize := 25.0;
               ysize := 25.0;
               xposition := 12.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'T': BEGIN
               plotter := tektronix;
               xunitspercm := 50.0;
               yunitspercm := 50.0;
               xsize := 20.46;
               ysize := 15.6;
               xposition := 10.23;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'H': BEGIN
               plotter := hp7470;
               penchange := yes;
               xunitspercm := 400.0;
               yunitspercm := 400.0;
               xsize := 24.0;
               ysize := 18.0;
               xposition := 12.89;
               yposition := 9.905;
               xoption := center;
               yoption := center;
               rotation := 0.0;
               END;
            'B': BEGIN
               plotter := bauschlomb;
               penchange := yes;
               xunitspercm := 100.0;
               yunitspercm := 100.0;
               xsize := 24.5;
               ysize := 17.5;
               xposition := 12.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'O': BEGIN
               plotter := other;
               (* initial parameter settings for a new
                  plotter go here *)
               END
            END;
      END; (* plotrparms *)

      PROCEDURE readparms;
         VAR parmnum : INTEGER;
            ch, c1, c2 : CHAR;

         PROCEDURE uppercase (VAR ch : CHAR);

            FUNCTION letter (ch : CHAR) : BOOLEAN;
            BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
               letter := ((ch >= 'A') AND (ch <= 'I'))
                         OR ((ch >= 'J') AND (ch <= 'R'))
                         OR ((ch >= 'S') AND (ch <= 'Z'))
                         OR ((ch >= 'a') AND (ch <= 'i'))
                         OR ((ch >= 'j') AND (ch <= 'r'))
                         OR ((ch >= 's') AND (ch <= 'z'));
            END; (* letter *)

        BEGIN (* uppercase *)
           IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                           OR (('A' > 'a') AND (ch < 'A')))
           THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
        END; (* uppercase *)

      PROCEDURE getp(VAR ch : CHAR);
         BEGIN (* getp *)
            ch := ' ';
            WHILE (ch=' ') AND NOT EOLN(parmfile) DO
               READ(parmfile, ch);
            uppercase(ch);
         END; (* getp *)

         FUNCTION setmode : modetype;
         BEGIN (* setmode *)
            getp(ch);
            CASE ch OF
               'Y': setmode := yes;
               'N': setmode := no
               END;
         END; (* setmode *)

      BEGIN (* readparms *)
         WHILE NOT EOF(parmfile) DO BEGIN
            c1 := ' ';
            c2 := ' ';
            WHILE NOT EOLN(parmfile) AND (c1=' ') DO
               READ(parmfile, c1);
            IF NOT EOLN(parmfile) THEN READ(parmfile, c2);
            REPEAT READ(parmfile, ch) UNTIL EOLN(parmfile)
                                                      OR (ch=' ');
            uppercase(c1);
            uppercase(c2);
            parmnum := 0;
            WHILE NOT((c1=ab1[parmnum]) AND (c2=ab2[parmnum])) AND
               (parmnum<=nabs) DO parmnum := parmnum+1;
            IF parmnum<=nabs
               THEN CASE parmnum OF
                  1: BEGIN
                     getp(ch);
                     plotrparms(ch)
                     END;
                  2: BEGIN
                     getp(ch);
                     CASE ch OF
                        'C': diagram := cladogram;
                        'E': BEGIN
                           diagram := phenogram;
                           vertical := levels;
                           horizontal := nearest;
                           drawing := rectangle;
                           widthmode := no
                           END;
                        'Y': BEGIN
                           diagram := phylogram;
                           vertical := lengths;
                           horizontal := nearest;
                           drawing := rectangle;
                           leveltips := no;
                           widthmode := no
                           END
                        END;
                     END;
                  3: BEGIN
                     getp(ch);
                     CASE ch OF
                        'V': vertical := levels;
                        'G': vertical := lengths;
                        'T': vertical := top;
                        'B': vertical := bottom;
                        'D': vertical := distance
                        END;
                     END;
                  4: BEGIN
                     getp(ch);
                     CASE ch OF
                        'S': horizontal := straightline;
                        'N': horizontal := nearest;
                        'F': horizontal := furthest
                        END;
                     END;
                  5: leveltips := setmode;
                  6: BEGIN
                     getp(ch);
                     CASE ch OF
                        'R': drawing := rectangle;
                        'S': drawing := slant
                        END;
                     END;
                  7: READ(parmfile, rotation);
                  8: BEGIN
                     getp(ch);
                     CASE ch OF
                        '1': fonttoload := font1;
                        '2': fonttoload := font2;
                        '3': fonttoload := font3;
                        '4': fonttoload := font4;
                        '5': fonttoload := font5
                        END;
                     END;
                  9: READ(parmfile, xunitspercm);
                  10: READ(parmfile, yunitspercm);
                  11: READ(parmfile, xsize);
                  12: READ(parmfile, ysize);
                  13: READ(parmfile, xposition);
                  14: READ(parmfile, yposition);
                  15: BEGIN
                     getp(ch);
                     CASE ch OF
                        'O': xoption := atroot;
                        'C': xoption := center;
                        'R': xoption := right;
                        'L': xoption := left
                        END;
                     END;
                  16: BEGIN
                     getp(ch);
                     CASE ch OF
                        'O': yoption := atroot;
                        'C': yoption := center;
                        'A': yoption := above;
                        'B': yoption := below
                        END;
                     END;
                  17: penchange := setmode;
                  18: widthmode := setmode;
                  19: READ (parmfile, treeheight);
                  20: READ (parmfile, treewidth);
                  21: READ (parmfile, labelheight);
                  22: READ (parmfile, spacing);
                  23: READ (parmfile, gap);
                  24: READ (parmfile, labelrotation);
                  25: popmode := setmode
                  END
               ELSE BEGIN
                  WRITELN(OUTPUT, ' PARAMETER LABEL THAT BEGINS ''',
                     c1, c2, ''' IS NOT A LEGITIMATE LABEL');
                  parmsok := FALSE;
                  READLN(parmfile);
                  END;
            READLN(parmfile);
            END;
      END; (* readparms *)

   BEGIN (* initparms *)
      ab1[0] := ' '; ab2[0] := ' ';
      ab1[1] := 'P'; ab2[1] := 'L'; ab1[2] := 'D'; ab2[2] := 'I';
      ab1[3] := 'V'; ab2[3] := 'E'; ab1[4] := 'H'; ab2[4] := 'O';
      ab1[5] := 'T'; ab2[5] := 'I'; ab1[6] := 'D'; ab2[6] := 'R';
      ab1[7] := 'R'; ab2[7] := 'O'; ab1[8] := 'F'; ab2[8] := 'O';
      ab1[9] := 'X'; ab2[9] := 'U'; ab1[10] := 'Y'; ab2[10] := 'U';
      ab1[11] := 'X'; ab2[11] := 'S'; ab1[12] := 'Y'; ab2[12] := 'S';
      ab1[13] := 'X'; ab2[13] := 'P'; ab1[14] := 'Y'; ab2[14] := 'P';
      ab1[15] := 'X'; ab2[15] := 'O'; ab1[16] := 'Y'; ab2[16] := 'O';
      ab1[17] := 'P'; ab2[17] := 'E'; ab1[18] := 'W'; ab2[18] := 'I';
      ab1[19] := 'T'; ab2[19] := 'H'; ab1[20] := 'T'; ab2[20] := 'W';
      ab1[21] := 'L'; ab2[21] := 'H'; ab1[22] := 'S'; ab2[22] := 'P';
      ab1[23] := 'G'; ab2[23] := 'A'; ab1[24] := 'L'; ab2[24] := 'R';
      ab1[25] := 'P'; ab2[25] := 'O';
      garbnode := NIL;
      garblabel := NIL;
      diagram := cladogram;
      vertical := distance;
      horizontal := straightline;
      leveltips := yes;
      popmode := no;
      drawing := slant;
      widthmode := yes;
      nnodes := 0;
      ntips := 0;
      nlabels := 0;
      pi := 3.1415926536;
      parmsok := TRUE;
      labelspresent := FALSE;
      fontloaded := FALSE;
      penchange := no;
      sizemode := yes;
      fonttoload := font1;
      treeheight := 12.5;
      treewidth := 25.5;
      labelrotation := 0.0;
      labelheight := 0.5;
      spacing := 1.0;
      gap := 1.0;
      plotrparms('H');  (* default plotter designation *)
      readparms;
      IF popmode = yes
         THEN popval := 1
         ELSE popval := 0;
      gap := 0.5 * gap * labelheight;
      spacing := 0.75 * spacing * labelheight;
      rotation := rotation * pi/180.0;
      labelrotation := rotation - pi/2.0 + labelrotation*pi/180.0;
      aspectratio := yunitspercm/xunitspercm;
   END; (* initparms *)

   PROCEDURE newnode(VAR p : nodeptr);
   BEGIN (* newnode *)
      IF garbnode <> NIL
         THEN BEGIN
            p := garbnode;
            garbnode := garbnode^.descendant;
            END
         ELSE NEW(p);
      nnodes := nnodes + 1;
      nodelist[nnodes] := p;
      p^.nodenum := nnodes;
      p^.ancestor := NIL;
      p^.descendant := NIL;
      p^.sibling := NIL;
      p^.lastdesc := NIL;
      p^.labels := NIL;
      p^.nodedone := FALSE;
      p^.popped := FALSE;
      p^.level := 0.0;
      p^.ystem := 0.0;
   END; (* newnode *)

   PROCEDURE newlabel(VAR lp : labelptr);
   BEGIN (* newlabel *)
      IF garblabel <> NIL
         THEN BEGIN
            lp := garblabel;
            garblabel := garblabel^.nextlab;
            END
         ELSE NEW(lp);
      nlabels := nlabels+1;
      labellist[nlabels] := lp;
      lp^.nextlab := NIL;
      lp^.nch := 0;
   END; (* newlabel *)

   PROCEDURE readtree;
      VAR i : INTEGER;
         ch : CHAR;
         node : nodeptr;

      PROCEDURE addnode(anc : nodeptr; VAR linktonode,
         anclastdesc : nodeptr);
         CONST blank = ' ';
            point = '.';
            comma = ',';
            colon = ':';
            leftp = '(';
            rightp = ')';
            aster = '*';
            pound = '#';
         VAR node : nodeptr;
            doingdescendants : BOOLEAN;

         PROCEDURE getch(VAR ch : CHAR);
         BEGIN (* getch *)
            IF EOF(treefile)
               THEN BEGIN
                  WRITELN(OUTPUT, '  NO ''*'' AT END OF TREE ',
                     'DESCRIPTION, ''*'' ASSUMED');
                  ch := '*';
                  END
               ELSE BEGIN
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  END;
         END; (* getch *)

         PROCEDURE processnodelabel(node : nodeptr);
            VAR lp, lp2 : labelptr;
         BEGIN (* processnodelabel *)
            newlabel(lp);
            IF node^.labels=NIL
               THEN BEGIN
                  labelspresent := TRUE;
                  node^.labels := lp;
                  END
               ELSE BEGIN
                  lp2 := node^.labels;
                  WHILE lp2^.nextlab <> NIL DO lp2 := lp2^.nextlab;
                  lp2^.nextlab := lp;
                  END;
            REPEAT
               IF lp^.nch < maxnch THEN BEGIN
                  lp^.nch := lp^.nch + 1;
                  IF ch = pound THEN ch := blank;
                  lp^.lab[lp^.nch] := ch;
                  getch(ch);
                  END;
               UNTIL (ch=blank) OR (ch=leftp) OR (ch=comma) OR
                  (ch=rightp) OR (ch=colon) OR (ch=aster);
         END; (* processnodelabel *)

         PROCEDURE processlength(node : nodeptr);
            VAR digit, ordzero, ordnine : INTEGER;
               value, divisor : REAL;
               pointread : BOOLEAN;
         BEGIN (* processlength *)
            ordzero := ORD('0');
            ordnine := ORD('9');
            pointread := FALSE;
            value := 0.0;
            divisor := 1.0;
            REPEAT getch(ch) UNTIL ch <> blank;
            digit := ORD(ch)-ordzero;
            WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point)
               DO BEGIN
               IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     value := value*10.0 + digit;
                     IF pointread THEN divisor := divisor*10.0;
                     END;
               getch(ch);
               digit := ORD(ch)-ordzero;
               END;
            node^.level := value/divisor;
         END; (* processlength *)

      BEGIN (* addnode *)
         doingdescendants := FALSE;
         newnode(node);
         linktonode := node;
         anclastdesc := node;
         node^.ancestor := anc;
         WHILE ch=blank DO getch(ch);
         WHILE (ch <> aster) AND (((ch <> comma) AND
            (ch <> rightp)) OR doingdescendants)
            DO BEGIN
            IF ch = leftp
               THEN BEGIN
                  doingdescendants := TRUE;
                  getch(ch);
                  addnode(node, node^.descendant,
                     node^.lastdesc);
                  END
               ELSE IF ch = comma
                  THEN BEGIN
                     getch(ch);
                     addnode(node, node^.lastdesc^.sibling,
                        node^.lastdesc);
                     END
                  ELSE IF ch = rightp
                     THEN BEGIN
                        doingdescendants := FALSE;
                        REPEAT getch(ch) UNTIL ch <> blank;
                        END
                     ELSE IF ch = colon
                        THEN processlength(node)
                        ELSE IF ch <> aster
                           THEN processnodelabel(node);
            WHILE ch = blank DO getch(ch);
            END;
      END; (* addnode *)

   BEGIN (* readtree *)
      ch := ' ';
      addnode(NIL, root, root);
      READLN(treefile);
   END; (* readtree *)

   PROCEDURE calctree(root : nodeptr);
      VAR count : INTEGER;
         fullscale, maxlevel, minlevel : REAL;

      PROCEDURE popall(p : nodeptr);
         VAR newp, prev : nodeptr;
            lp : labelptr;
      BEGIN (* popall *)
         prev := NIL;
         lp := p^.labels;
         WHILE lp <> NIL DO BEGIN
            newnode(newp);
            newp^.popped := TRUE;
            lp := lp^.nextlab;
            IF p^.descendant = NIL
               THEN p^.lastdesc := newp
               ELSE newp^.sibling := p^.descendant;
            p^.descendant := newp;
            newp^.ancestor := p;
            IF vertical = levels THEN newp^.level := p^.level;
            END;
         lp := p^.labels;
         p^.labels := NIL;
         newp := p^.descendant;
         WHILE lp <> NIL DO BEGIN
            newp^.labels := lp;
            lp := lp^.nextlab;
            newp^.labels^.nextlab := NIL;
            newp := newp^.sibling;
            END;
      END; (* popall *)

      PROCEDURE popout(p : nodeptr);
      BEGIN (* popout *)
         IF p^.descendant <> NIL
            THEN BEGIN
               popout(p^.descendant);
               IF p^.labels <> NIL THEN popall(p);
               END
            ELSE IF p^.labels^.nextlab <> NIL
                  THEN popall(p);
         IF p^.sibling <> NIL THEN popout(p^.sibling);
      END; (* popout *)

      PROCEDURE counttips(p : nodeptr);
         VAR ptr : nodeptr;
      BEGIN (* counttips *)
         p^.fromtop := 0;
         IF p = root
            THEN p^.frombottom := 0
            ELSE BEGIN
               IF p^.popped
                  THEN p^.frombottom := p^.ancestor^.frombottom +
                                                              popval
                  ELSE p^.frombottom := p^.ancestor^.frombottom + 1;
               END;
         IF p^.frombottom > maxdepth THEN maxdepth := p^.frombottom;
         IF p^.descendant = NIL
            THEN BEGIN
               ntips := ntips + 1;
               IF p^.labels<>NIL THEN p^.labels^.tipnum := ntips;
               END
            ELSE BEGIN
               counttips(p^.descendant);
               ptr := p^.descendant;
               WHILE ptr <> NIL DO BEGIN
                  IF ptr^.popped
                     THEN BEGIN
                        IF p^.fromtop < ptr^.fromtop+popval
                           THEN p^.fromtop := ptr^.fromtop+popval;
                        END
                     ELSE BEGIN
                        IF ptr^.fromtop+1 > p^.fromtop
                           THEN p^.fromtop := ptr^.fromtop+1;
                        END;
                  ptr := ptr^.sibling;
                  END;
               ptr := p^.descendant;
               WHILE ptr <> NIL DO BEGIN
                  IF ptr^.popped THEN ptr^.fromtop := p^.fromtop-popval;
                  ptr := ptr^.sibling;
                  END;
               END;
         IF p^.sibling <> NIL THEN counttips(p^.sibling);
      END; (* counttips *)
 PROCEDURE placenodes(p : nodeptr; VAR count : INTEGER);
      BEGIN (* placenodes *)
         IF p^.descendant = NIL
            THEN BEGIN
               p^.left := p;
               p^.right := p;
               count := count+1;
               p^.x := (count-1)/(ntips-1);
               CASE vertical OF
                  top: p^.level := 1.0-p^.fromtop/maxdepth;
                  bottom: p^.level := p^.frombottom/maxdepth;
                  distance: p^.level := 0.0
                  END;
               p^.nodedone := TRUE;
               END
            ELSE BEGIN
               placenodes(p^.descendant, count);
               p^.left := p^.descendant^.left;
               p^.right := p^.lastdesc^.right;
               IF horizontal = furthest
                  THEN p^.x := (p^.left^.x+p^.right^.x)/2.0
                  ELSE p^.x := (p^.descendant^.x+p^.lastdesc^.x)/2.0;
               CASE vertical OF
                  top: p^.level := 1.0-p^.fromtop/maxdepth;
                  bottom: p^.level := p^.frombottom/maxdepth;
                  distance: p^.level := -(p^.right^.x-p^.left^.x)
                  END;
               END;
         IF p^.sibling <> NIL THEN placenodes(p^.sibling, count);
      END; (* placenodes *)

      PROCEDURE addlengths(p : nodeptr);
         VAR currentlevel : REAL;
      BEGIN (* addlengths *)
         IF p <> root THEN p^.level := p^.level + p^.ancestor^.level;
         IF p^.descendant <> NIL THEN addlengths(p^.descendant);
         IF p^.sibling <> NIL THEN addlengths(p^.sibling);
      END; (* addlengths *)

      PROCEDURE lineupnodes(p : nodeptr);

         PROCEDURE singleline(p : nodeptr; side : sidetype;
               pstart, pend : nodeptr);
         BEGIN (* singleline *)
            WHILE p <> NIL DO BEGIN
               p^.x := (p^.y-pstart^.y)*(pend^.x-pstart^.x)/
                                  (pend^.y-pstart^.y)+pstart^.x;
               p^.nodedone := TRUE;
               IF side = leftside
                  THEN p := p^.descendant
                  ELSE p := p^.lastdesc;
               END;
         END; (* singleline *)

      BEGIN (* lineupnodes *)
         IF p <> NIL THEN BEGIN
            IF NOT p^.nodedone THEN BEGIN
               IF ABS(p^.ancestor^.x-p^.right^.x) >=
                       ABS(p^.ancestor^.x-p^.left^.x)
                  THEN singleline(p, rightside, p^.ancestor,
                     p^.right)
                  ELSE singleline(p, leftside, p^.ancestor,
                     p^.left);
               END;
            lineupnodes(p^.descendant);
            lineupnodes(p^.sibling);
            END;
      END; (* lineupnodes *)

      PROCEDURE levellimits(p : nodeptr);
      BEGIN (* levellimits *)
         IF p^.descendant <> NIL THEN levellimits(p^.descendant);
         IF p^.sibling <> NIL THEN levellimits(p^.sibling);
         IF p^.level > maxlevel THEN maxlevel := p^.level;
         IF p^.level < minlevel THEN minlevel := p^.level;
      END; (* levellimits *)

      PROCEDURE scalenodes(p : nodeptr);
      BEGIN (* scalenodes *)
         IF p^.descendant <> NIL
               THEN scalenodes(p^.descendant);
         IF (leveltips = yes) AND (p^.descendant = NIL)
            THEN p^.y := 1.0
            ELSE p^.y := ABS(p^.level-root^.level)/fullscale;
         IF p^.labels<>NIL THEN p^.labels^.ylab := p^.y;
         IF p^.sibling <> NIL THEN scalenodes(p^.sibling);
      END; (* scalenodes *)

      PROCEDURE setstems(p : nodeptr);
         VAR ptr : nodeptr;
      BEGIN (* setstems *)
         IF p^.descendant <> NIL THEN setstems(p^.descendant);
         p^.xstem := p^.x;
         ptr := p^.descendant;
         WHILE ptr <> NIL DO BEGIN
            ptr^.ystem := p^.y;
            ptr := ptr^.sibling;
            END;
         IF p^.sibling <> NIL THEN setstems(p^.sibling);
      END; (* setstems *)

   BEGIN (* calctree *)
      popout(root);
      counttips(root);
      count := 0;
      placenodes(root, count);
      IF vertical = lengths THEN addlengths(root);
      maxlevel := -bignum;
      minlevel := bignum;
      levellimits(root);
      fullscale := maxlevel-minlevel;
      scalenodes(root);
      IF horizontal = straightline THEN BEGIN
         root^.x := 0.5;
         root^.nodedone := TRUE;
         lineupnodes(root);
         END;
      IF drawing = rectangle THEN setstems(root);
   END; (* calctree *)

   PROCEDURE hardwarecoordinates;
      VAR i : INTEGER;
         factor, xmax, xmin, ymax, ymin, thislength,
            dsin, dcos, lsin, lcos, xtrans, ytrans,
            xcorrect, ycorrect, dummy : REAL;
         lp : labelptr;
         p : nodeptr;

      FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER;
         height, compress : REAL; VAR font : fonttype) : REAL;
         VAR i, j, code, coord : INTEGER;
            sumlength, heightfont, widthfont, yfactor : REAL;
      BEGIN (* lengthext *)
         sumlength := 0.0;
         FOR i := 1 TO nchars DO BEGIN
            code := ORD(pstring[i]);
            j := 1;
            WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
               j := font[j];
            IF font[j+1] = code THEN BEGIN
               heightfont := font[j+2];
               widthfont := font[j+3];
               sumlength := sumlength + (height/heightfont)
                     *widthfont;
               END;
            END;
            lengthtext := sumlength*compress;
      END; (* lengthtext *)

      PROCEDURE rotate(VAR x, y : REAL; xdisp, ydisp, xcenter, ycenter,
             sinrot, cosrot : REAL);
      BEGIN (* rotate *)
         x := xcenter+xdisp*cosrot+ydisp*sinrot;
         IF xmin>x THEN xmin := x;
         IF xmax<x THEN xmax := x;
         y := ycenter-xdisp*sinrot+ydisp*cosrot;
         IF ymin>y THEN ymin := y;
         IF ymax<y THEN ymax := y;
      END; (* rotate *)

   BEGIN (* hardwarecoordinates *)
      xmin := bignum;
      xmax := -bignum;
      ymin := bignum;
      ymax := -bignum;
      dsin := SIN(rotation);
      dcos := COS(rotation);
      lsin := SIN(labelrotation);
      lcos := COS(labelrotation);
      IF widthmode = yes
         THEN BEGIN
            factor := treewidth / ((ntips-1)*(labelheight+spacing));
            IF factor < 1.0
               THEN BEGIN
                  labelheight := labelheight*factor;
                  spacing := spacing*factor;
                  END
               ELSE spacing := (treewidth/(ntips-1)) - labelheight;
            END
         ELSE treewidth := (ntips-1)*(labelheight+spacing);
      xcorrect := -labelheight*(1+SIN(labelrotation-rotation)/2.0);
      ycorrect := -labelheight*COS(labelrotation-rotation)/2.0;
      FOR i := 1 TO nlabels DO BEGIN
         lp := labellist[i];
         lp^.xlab := lp^.tipnum*labelheight +
                           (lp^.tipnum-1)*spacing + xcorrect;
         lp^.ylab := lp^.ylab*treeheight + gap + ycorrect;
         thislength := lengthtext(lp^.lab, lp^.nch, labelheight,
               1.0, font);
         rotate(lp^.xlab,lp^.ylab,lp^.xlab,lp^.ylab,0.0,0.0,dsin,dcos);
         rotate(dummy,dummy,0.0,labelheight,lp^.xlab,lp^.ylab,
               lsin,lcos);
         rotate(dummy,dummy,thislength,labelheight,lp^.xlab,lp^.ylab,
               lsin,lcos);
         rotate(dummy,dummy,thislength,0.0,lp^.xlab,lp^.ylab,lsin,lcos);
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := p^.x * treewidth;
         p^.y := p^.y * treeheight;
         rotate(p^.x,p^.y,p^.x,p^.y,0.0,0.0,dsin,dcos);
         IF drawing = rectangle THEN BEGIN
            p^.xstem := p^.xstem * treewidth;
            p^.ystem := p^.ystem * treeheight;
            rotate(p^.xstem,p^.ystem,p^.xstem,p^.ystem,0.0,0.0,
                  dsin,dcos);
            END;
         END;
      xrange := xmax - xmin;
      yrange := ymax - ymin;
      IF sizemode = yes
         THEN BEGIN
            IF ysize/xsize > yrange/xrange
               THEN factor := xsize/xrange
               ELSE factor := ysize/yrange;
            END
         ELSE factor := 1.0;
      CASE xoption OF
         atroot: BEGIN END;
         right: xtrans := xposition - xmin*factor;
         left: xtrans := xposition - xmax*factor;
         center: xtrans := xposition + (xrange/2.0 - xmax)*factor
         END;
      CASE yoption OF
         atroot: BEGIN END;
         above: ytrans := yposition - ymin*factor;
         below: ytrans := yposition - ymax*factor;
         center: ytrans := yposition + (yrange/2.0 - ymax)*factor
         END;
      FOR i := 1 TO nlabels DO BEGIN
         lp := labellist[i];
         lp^.xlab := (lp^.xlab*factor+xtrans)*xunitspercm;
         lp^.ylab := (lp^.ylab*factor+ytrans)*xunitspercm;
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := (p^.x*factor+xtrans)*xunitspercm;
         p^.y := (p^.y*factor+ytrans)*xunitspercm;
         IF (drawing = rectangle) AND (i <> 1) THEN BEGIN
            p^.xstem := (p^.xstem*factor+xtrans)*xunitspercm;
            p^.ystem := (p^.ystem*factor+ytrans)*xunitspercm;
            END;
         END;
      labelheight := labelheight * factor * xunitspercm;
      xrange := xrange * factor * xunitspercm;
      yrange := yrange * factor *