PROGRAM Plotree(parmfile, treefile, fontfile, plotfile, OUTPUT);
   (* PHYLIP version 2.7
      Version 1.1, May 1985.  Written by Chris Meacham, Botany Dept.,
      Univ. of Georgia, Athens, Georgia 30602, U.S.A. *)

   CONST bignum = 1000000.0;
      maxnodes = 100;
      maxlablen = 3;
      fontsize = 2905;  (* Required for font1: 1035; font2: 1885;  *)
      doswitch = TRUE;  (*   font3: 2905; font4: 2012; font5: 2902 *)
      dontswitch = FALSE;

   TYPE fontname = (font1, font2, font3, font4, font5);
      modetype = (yes, no);
      posoptiontype = (atroot, right, left, above, below, center);
      penstatustype = (penup, pendown);
      plottertype = (calcomp, tektronix, hp7470, bauschlomb, other);
      fonttype = ARRAY [1..fontsize] OF INTEGER;
      plotstring = ARRAY [1..maxlablen] OF CHAR;
      nodeptr = ^treenode;
      treenode = RECORD
         ancestor, sibling, descendant, lastdesc,
            left, right : nodeptr;
         c, nbranches, nch : INTEGER;
         nodesize, len, spread, r, theta, x, y,
         height, width : REAL;
         ndlabel : plotstring
         END;
   VAR nnodes, xnow, ynow, oldxhigh, oldxlow, oldyhigh,
         oldylow, longestlablen : INTEGER;
      pi, smallnode, radius, absradius, diameter, spanunit,
         padding, spaceval, scrunchval, heightval,
         scrunchlimit, rotation, xposition, yposition,
         xunitspercm, yunitspercm, aspectratio, 
         xsize, ysize, xrange, yrange, maxdiam, labelheight,
         labelsize : REAL;
      xcircle, ycircle : ARRAY [1..30] OF REAL;
      cchex : ARRAY [0..15] OF CHAR;
      parmfile, treefile, fontfile, plotfile : TEXT;
      labelspresent, fontloaded, candraw, parmsok : BOOLEAN;
      fonttoload : fontname;
      font : fonttype;
      sizemode, lengthmode, monitor, penchange : modetype;
      plotter : plottertype;
      xoption, yoption : posoptiontype;
      root, garbage : nodeptr;
      nodelist : ARRAY [1..maxnodes] OF nodeptr;

   PROCEDURE initparms;
      CONST nabs = 21;
      VAR ab1, ab2 : ARRAY [0..nabs] OF CHAR;

      PROCEDURE plotrparms(ch : CHAR);
      BEGIN (* plotrparms *)
         CASE ch OF
            'C': BEGIN
               plotter := calcomp;
               xunitspercm := 39.37;
               yunitspercm := 39.37;
               xsize := 25.0;
               ysize := 25.0;
               xposition := 12.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'T': BEGIN
               plotter := tektronix;
               xunitspercm := 50.0;
               yunitspercm := 50.0;
               xsize := 20.46;
               ysize := 15.6;
               xposition := 10.23;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'H': BEGIN
               plotter := hp7470;
               penchange := yes;
               xunitspercm := 400.0;
               yunitspercm := 400.0;
               xsize := 24.0;
               ysize := 18.0;
               xposition := 12.85;
               yposition := 9.55;
               xoption := center;
               yoption := center;
               rotation := 90.0;
               END;
            'B': BEGIN
               plotter := bauschlomb;
               penchange := yes;
               xunitspercm := 100.0;
               yunitspercm := 100.0;
               xsize := 24.5;
               ysize := 17.5;
               xposition := 12.5;
               yposition := 0.0;
               xoption := center;
               yoption := above;
               rotation := 0.0;
               END;
            'O': BEGIN
               plotter := other;
               (* initial parameter settings for a new
                  plotter go here *)
               END
            END;
         END; (* plotrparms *)

      PROCEDURE readparms;
         VAR parmnum : INTEGER;
            ch, c1, c2 : CHAR;

         PROCEDURE uppercase (VAR ch : CHAR);

            FUNCTION letter (ch : CHAR) : BOOLEAN;
            BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
               letter := ((ch >= 'A') AND (ch <= 'I'))
                         OR ((ch >= 'J') AND (ch <= 'R'))
                         OR ((ch >= 'S') AND (ch <= 'Z'))
                         OR ((ch >= 'a') AND (ch <= 'i'))
                         OR ((ch >= 'j') AND (ch <= 'r'))
                         OR ((ch >= 's') AND (ch <= 'z'));
               END; (* letter *)

         BEGIN (* uppercase *)
            IF letter(ch) AND ((('A' < 'a') AND (ch >= 'a'))
                            OR (('A' > 'a') AND (ch < 'A')))
            THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
            END; (* uppercase *)

         PROCEDURE getp(VAR ch : CHAR);
         BEGIN (* getp *)
            ch := ' ';
            WHILE (ch=' ') AND NOT EOLN(parmfile) DO
               READ(parmfile, ch);
            uppercase(ch);
            END; (* getp *)

         FUNCTION setmode : modetype;
         BEGIN (* setmode *)
            getp(ch);
            CASE ch OF
               'Y': setmode := yes;
               'N': setmode := no
               END;
            END; (* setmode *)

      BEGIN (* readparms *)
         WHILE NOT EOF(parmfile) DO BEGIN
            c1 := ' ';
            c2 := ' ';
            WHILE NOT EOLN(parmfile) AND (c1=' ') DO
               READ(parmfile, c1);
            IF NOT EOLN(parmfile) THEN READ(parmfile, c2);
            REPEAT READ(parmfile, ch) UNTIL EOLN(parmfile) OR (ch=' ');
            uppercase(c1);
            uppercase(c2);
            parmnum := 0;
            WHILE NOT((c1=ab1[parmnum]) AND (c2=ab2[parmnum])) AND
               (parmnum<=nabs) DO parmnum := parmnum+1;
            IF parmnum <= nabs
               THEN CASE parmnum OF
                  1: BEGIN
                     getp(ch);
                     plotrparms(ch);
                     END;
                  2: lengthmode := setmode;
                  3: sizemode := setmode;
                  4: READ(parmfile, diameter);
                  5: READ(parmfile, spanunit);
                  6: READ(parmfile, smallnode);
                  7: READ(parmfile, rotation);
                  8: BEGIN
                     getp(ch);
                     CASE ch OF
                       '1': fonttoload := font1;
                       '2': fonttoload := font2;
                       '3': fonttoload := font3;
                       '4': fonttoload := font4;
                       '5': fonttoload := font5
                        END;
                     END;
                  9: READ(parmfile, xunitspercm);
                  10: READ(parmfile, yunitspercm);
                  11: READ(parmfile, xsize);
                  12: READ(parmfile, ysize);
                  13: READ(parmfile, xposition);
                  14: READ(parmfile, yposition);
                  15: BEGIN
                     getp(ch);
                     CASE ch OF
                        'O': xoption := atroot;
                        'C': xoption := center;
                        'R': xoption := right;
                        'L': xoption := left
                        END;
                     END;
                  16: BEGIN
                     getp(ch);
                     CASE ch OF
                        'O': yoption := atroot;
                        'C': yoption := center;
                        'A': yoption := above;
                        'B': yoption := below
                        END;
                     END;
                  17: READ(parmfile, padding);
                  18: READ(parmfile, maxdiam);
                  19: monitor := setmode;
                  20: penchange := setmode;
                  21: READ(parmfile, labelsize)
                  END
               ELSE BEGIN
                  WRITELN(OUTPUT, ' PARAMETER LABEL THAT BEGINS ''',
                     c1, c2, ''' IS NOT A LEGITIMATE LABEL');
                  parmsok := FALSE;
                  READLN(parmfile);
                  END;
            READLN(parmfile);
            END;
         END; (* readparms *)

   BEGIN (* initparms *)
      ab1[0] := ' '; ab2[0] := ' ';
      ab1[1] := 'P'; ab2[1] := 'L'; ab1[2] := 'L'; ab2[2] := 'E';
      ab1[3] := 'S'; ab2[3] := 'I'; ab1[4] := 'N'; ab2[4] := 'O';
      ab1[5] := 'I'; ab2[5] := 'N'; ab1[6] := 'S'; ab2[6] := 'M';
      ab1[7] := 'R'; ab2[7] := 'O'; ab1[8] := 'F'; ab2[8] := 'O';
      ab1[9] := 'X'; ab2[9] := 'U'; ab1[10] := 'Y'; ab2[10] := 'U';
      ab1[11] := 'X'; ab2[11] := 'S'; ab1[12] := 'Y'; ab2[12] := 'S';
      ab1[13] := 'X'; ab2[13] := 'P'; ab1[14] := 'Y'; ab2[14] := 'P';
      ab1[15] := 'X'; ab2[15] := 'O'; ab1[16] := 'Y'; ab2[16] := 'O';
      ab1[17] := 'P'; ab2[17] := 'A'; ab1[18] := 'M'; ab2[18] := 'A';
      ab1[19] := 'M'; ab2[19] := 'O'; ab1[20] := 'P'; ab2[20] := 'E';
      ab1[21] := 'L'; ab2[21] := 'A';
      garbage := NIL;
      pi := 3.1415926536;
      heightval := 0.9;
      scrunchval := 0.8;
      scrunchlimit := 0.01;
      padding := 1.1;
      parmsok := TRUE;
      labelspresent := FALSE;
      fontloaded := FALSE;
      monitor := no;
      penchange := no;
      lengthmode := no;
      sizemode := yes;
      fonttoload := font1;
      labelsize := 1.0;
      maxdiam := 4.0;
      smallnode := 0.6;
      spanunit := 3.0;
      diameter := 1.0;
      plotrparms('H');    (* default plotter designation *)
      readparms;
      padding := padding*diameter;
      IF (SQR(spanunit)-SQR(padding/2.0)) < 0.0
         THEN BEGIN
            WRITELN(OUTPUT, ' INTERNODE LENGTH IS TOO SMALL IN ',
               'RELATION TO NODE DIAMETER');
            parmsok := FALSE;
            END
         ELSE spaceval := 2.0*ARCTAN((padding/2.0)/SQRT(SQR(spanunit)-
            SQR(padding/2.0)));
      radius := diameter/2.0;
      labelsize := labelsize * 0.9;
      rotation := rotation * pi/180.0;
      aspectratio := yunitspercm/xunitspercm;
      END; (* initparms *)

   PROCEDURE initnodes(p, bottom : nodeptr);
   BEGIN (* initnodes *)
      IF p <> NIL
      THEN BEGIN
         initnodes(p^.descendant, bottom);
         IF p <> bottom THEN initnodes(p^.sibling, bottom);
         p^.left := p;
         p^.right := p;
         p^.height := 0.0;
         p^.width := 0.0;
         p^.r := 0.0;
         p^.theta := 0.0;
         END;
      END; (* initnodes *)

   PROCEDURE readtree;
      VAR i, count : INTEGER;
         ch : CHAR;
         node : nodeptr;

      PROCEDURE newnode(VAR p : nodeptr);
      BEGIN (* newnode *)
         IF garbage <> NIL
            THEN BEGIN
               p := garbage;
               garbage := garbage^.descendant;
               END
            ELSE NEW(p);
         count := count+1;
         nodelist[count] := p;
         p^.c := count;
         p^.ancestor := NIL;
         p^.descendant := NIL;
         p^.sibling := NIL;
         p^.lastdesc := NIL;
         p^.nch := 0;
         p^.nbranches := 0;
         p^.nodesize := 1.0;
         p^.len := spanunit;
         p^.spread := pi;
         END; (* newnode *)

      PROCEDURE addnode(anc : nodeptr; VAR linktonode,
         anclastdesc : nodeptr);
         CONST blank = ' ';
            point = '.';
            comma = ',';
            colon = ':';
            leftp = '(';
            rightp = ')';
            aster = '*';
            pound = '#';
         VAR node : nodeptr;
            doingdescendants : BOOLEAN;

         PROCEDURE getch(VAR ch : CHAR);
         BEGIN (* getch *)
            IF EOF(treefile)
               THEN BEGIN
                  WRITELN(OUTPUT, '  NO ''*'' AT END OF TREE ',
                     'DESCRIPTION, ''*'' ASSUMED');
                  ch := '*';
                  END
               ELSE BEGIN
                  IF EOLN(treefile) THEN READLN(treefile);
                  READ(treefile, ch);
                  END;
            END; (* getch *)

         PROCEDURE processnodelabel(node : nodeptr);
         BEGIN (* processnodelabel *)
            IF node^.nch=0
               THEN BEGIN
                  labelspresent := TRUE;
                  REPEAT
                     IF node^.nch<maxlablen THEN BEGIN
                        node^.nch := node^.nch+1;
                        IF ch = pound THEN ch := blank;
                        node^.ndlabel[node^.nch] := ch;
                        END;
                     getch(ch);
                     UNTIL (ch=blank) OR (ch=leftp) OR (ch=comma) OR
                        (ch=rightp) OR (ch=colon) OR (ch=aster);
                  END
               ELSE REPEAT getch(ch) UNTIL (ch=blank) OR (ch=leftp)
                  OR (ch=comma) OR (ch=rightp) OR (ch=colon) OR
                  (ch=aster);
            END; (* processnodelabel *)

         PROCEDURE processlength(node : nodeptr);
            VAR digit, ordzero : INTEGER;
               value, divisor : REAL;
               pointread : BOOLEAN;
         BEGIN (* processlength *)
            ordzero := ORD('0');
            pointread := FALSE;
            value := 0.0;
            divisor := 1.0;
            REPEAT getch(ch) UNTIL ch <> blank;
            digit := ORD(ch)-ordzero;
            WHILE ((digit >= 0) AND (digit <= 9)) OR (ch=point)
               DO BEGIN
               IF ch = point
                  THEN pointread := TRUE
                  ELSE BEGIN
                     value := value*10.0 + digit;
                     IF pointread THEN divisor := divisor*10.0;
                     END;
               getch(ch);
               digit := ORD(ch)-ordzero;
               END;
            IF lengthmode = yes 
               THEN node^.len := (value/divisor)*spanunit;
            END; (* processlength *)

      BEGIN (* addnode *)
         doingdescendants := FALSE;
         newnode(node);
         linktonode := node;
         anclastdesc := node;
         node^.ancestor := anc;
         WHILE ch=blank DO getch(ch);
         WHILE (ch <> aster) AND (((ch <> comma) AND
            (ch <> rightp)) OR doingdescendants) DO BEGIN
            IF ch = leftp
               THEN BEGIN
                  doingdescendants := TRUE;
                  getch(ch);
                  addnode(node, node^.descendant,
                     node^.lastdesc);
                  node^.nbranches := node^.nbranches+1;
                  END
               ELSE IF ch = comma
                  THEN BEGIN
                     getch(ch);
                     addnode(node, node^.lastdesc^.sibling,
                        node^.lastdesc);
                     node^.nbranches := node^.nbranches+1;
                     END
                  ELSE IF ch = rightp
                     THEN BEGIN
                        doingdescendants := FALSE;
                        REPEAT getch(ch) UNTIL ch <> blank;
                        END
                     ELSE IF ch = colon
                        THEN processlength(node)
                        ELSE IF ch <> aster
                           THEN processnodelabel(node);
            WHILE ch = blank DO getch(ch);
            END;
         END; (* addnode *)

   BEGIN (* readtree *)
      count := 0;
      ch := ' ';
      addnode(NIL, root, root);
      READLN(treefile);
      nnodes := count;
      initnodes(root, root);
      FOR i := 1 TO nnodes DO BEGIN
         node := nodelist[i];
         IF node^.nch = 0 THEN node^.nodesize := smallnode;
         IF (node^.nch = 1) AND (node^.ndlabel[1] = '@') THEN
            node^.nch := 0;
         END;
      IF lengthmode = yes THEN
         FOR i := 1 TO nnodes DO BEGIN
            node := nodelist[i];
            IF node <> root THEN node^.len := node^.len +
               (node^.nodesize + node^.ancestor^.nodesize)*radius;
            END;
      END; (* readtree *)

   PROCEDURE calctree(p : nodeptr);
      VAR fitok : BOOLEAN;
         current : nodeptr;

         PROCEDURE fanflip(p, bottom : nodeptr; angle : REAL;
            flip : BOOLEAN);
            VAR current : nodeptr;
         BEGIN (* fanflip *)
            IF p <> NIL THEN BEGIN
               fanflip(p^.descendant, bottom, angle, flip);
               IF p <> bottom THEN fanflip(p^.sibling, bottom,
                                                 angle, flip);
               IF flip
                  THEN BEGIN
                     p^.theta := angle-p^.theta;
                     current := p^.descendant;
                     p^.descendant := p^.lastdesc;
                     WHILE current <> p^.descendant DO BEGIN
                        p^.lastdesc^.sibling := current;
                        p^.lastdesc := p^.lastdesc^.sibling;
                        current := current^.sibling;
                        p^.lastdesc^.sibling := NIL;
                        END;
                     current := p^.left;
                     p^.left := p^.right;
                     p^.right := current;
                     END
                  ELSE p^.theta := angle + p^.theta;
               IF p^.theta < bottom^.ancestor^.left^.theta
                  THEN bottom^.ancestor^.left := p;
               IF p^.theta > bottom^.ancestor^.right^.theta
                  THEN bottom^.ancestor^.right := p;
               END;
            END; (* fanflip *)

      PROCEDURE arrange(p : nodeptr; VAR fitok : BOOLEAN);
         VAR nlong : INTEGER;
            maxheight, angle, angleused, spacing : REAL;
            flip : BOOLEAN;
            previous, current, next : nodeptr;

         PROCEDURE switchbranches(VAR ance, prev, curr, next
               : nodeptr;
            switch : BOOLEAN);
         BEGIN (* switchbranches *)
            IF switch
               THEN BEGIN
                  IF prev = ance
                     THEN ance^.descendant := next
                     ELSE prev^.sibling := next;
                  curr^.sibling := next^.sibling;
                  next^.sibling := curr;
                  prev := ance;
                  curr := ance^.descendant;
                  next := curr^.sibling;
                  END
               ELSE BEGIN
                  prev := curr;
                  curr := next;
                  next := next^.sibling;
                  END;
            END; (* switchbranches *)

      BEGIN (* arrange *)
         previous := p;
         current := p^.descendant;
         next := current^.sibling;
         WHILE next <> NIL DO
            IF (current^.height < next^.height) OR
               ((current^.height = next^.height) AND
               (current^.width < next^.width))
               THEN switchbranches(p, previous, current, next, doswitch)
               ELSE switchbranches(p, previous, current,
                                                   next, dontswitch);
         nlong := 0;
         current := p^.descendant;
         maxheight := current^.height;
         WHILE current <> NIL DO BEGIN
            IF current^.height >= heightval*maxheight
               THEN nlong := nlong + 1;
            current := current^.sibling;
            END;
         angleused := 0.0;
         current := p^.descendant;
         WHILE current <> NIL DO BEGIN
            angleused := angleused + current^.width;
            current := current^.sibling;
            END;
         spacing := (p^.spread-angleused)/(p^.nbranches+1);
         fitok := spacing >= spaceval;
         current := p^.descendant;
         IF ODD(nlong)
            THEN angle := 0.0
            ELSE angle := spacing/2.0 + current^.right^.theta;
         p^.right := current^.left;
         p^.left := current^.right;
         flip := TRUE;
         fanflip(current, current, angle, flip);
         current := current^.sibling;
         WHILE current <> NIL DO BEGIN
            IF ABS(p^.left^.theta) > ABS(p^.right^.theta)
               THEN angle := p^.right^.theta
                             + spacing + current^.right^.theta
               ELSE angle := p^.left^.theta - spacing
                             - current^.right^.theta;
            flip := angle >= 0.0;
            fanflip(current, current, angle, flip);
            current := current^.sibling;
            END;
         previous := p;
         current := p^.descendant;
         next := current^.sibling;
         WHILE next <> NIL DO
            IF current^.theta > next^.theta
               THEN switchbranches(p, previous, current, next, doswitch)
               ELSE switchbranches(p, previous, current, next,
                                                     dontswitch);
         p^.lastdesc := current;
         END; (* arrange *)
urrent, current, angle, flip);
         current := current^.sibling;
         WHILE current <> NIL DO BEGIN
            IF ABS(p^.left^.theta) > ABS(p^.right^.theta)
               THEN angle := p^.right^.theta
                             + spacing +