
      PROCEDURE scrunch(p, bottom : nodeptr);
      BEGIN (* scrunch *)
         IF p <> NIL THEN BEGIN
            IF (monitor = yes) AND (p = bottom)
               THEN WRITELN(OUTPUT, '    SCRUNCHING AT NODE', p^.c:4);
            scrunch(p^.descendant, bottom);
            IF p <> bottom THEN scrunch(p^.sibling, bottom);
            p^.spread := p^.spread * scrunchval;
            IF p^.spread < scrunchlimit THEN candraw := FALSE;
            END;
         END; (* scrunch *)

      PROCEDURE polartrans(p, bottom : nodeptr; len : REAL);
         VAR pr, ptheta : REAL;
      BEGIN (* polartrans *)
         IF p <> NIL THEN BEGIN
            polartrans(p^.descendant, bottom, len);
            IF p <> bottom THEN polartrans(p^.sibling, bottom, len);
            pr := p^.r;
            ptheta := p^.theta;
            p^.r := SQRT(SQR(len) + SQR(pr)+ 2*len*pr*COS(ptheta));
            p^.theta := ARCTAN(pr*SIN(ptheta)/(pr*COS(ptheta) + len));
            IF p^.r  > bottom^.height THEN bottom^.height := p^.r;
            IF p^.theta > bottom^.right^.theta THEN bottom^.right := p;
            IF p^.theta < bottom^.left^.theta THEN bottom^.left := p;
            END;
         END; (* polartrans *)

   BEGIN (* calctree *)
      IF (p <> NIL) AND candraw THEN BEGIN
         IF monitor = yes THEN WRITELN(OUTPUT, '     AT NODE',p^.c:4);
         REPEAT
            calctree(p^.descendant);
            IF p^.descendant <> NIL
               THEN arrange(p,fitok)
               ELSE fitok := TRUE;
            IF fitok
               THEN BEGIN
                  IF p <> root THEN BEGIN
                     p^.right := p;
                     p^.left := p;
                     polartrans(p, p, p^.len);
                     p^.width := p^.right^.theta - p^.left^.theta;
                     END;
                  END
               ELSE BEGIN
                  current := p^.descendant;
                  WHILE current <> NIL DO BEGIN
                     scrunch(current, current);
                     initnodes(current, current);
                     current := current^.sibling;
                     END;
                  END;
            UNTIL fitok OR NOT candraw;
         IF candraw THEN BEGIN
            IF ABS(p^.right^.theta) > ABS(p^.left^.theta)
               THEN fanflip(p, p, 0.0, TRUE);
            calctree(p^.sibling);
            END;
         END;
      END; (* calctree *)

   PROCEDURE hardwarecoordinates;
      VAR i : INTEGER;
         factor, angle, xmax, xmin, ymax, ymin, thisx, thisy,
         thisrad, xtrans, ytrans : REAL;
         p : nodeptr;
   BEGIN (* hardwarecoordinates *)
      IF rotation <> 0.0 THEN BEGIN
         FOR i := 1 TO nnodes DO
            nodelist[i]^.theta := nodelist[i]^.theta + rotation;
         END;
      xmax := -bignum;
      xmin := bignum;
      ymax := -bignum;
      ymin := bignum;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := p^.r*SIN(p^.theta);
         p^.y := p^.r*COS(p^.theta);
         thisx := p^.x;
         thisy := p^.y;
         thisrad := p^.nodesize*radius;
         IF thisx + thisrad > xmax THEN xmax := thisx + thisrad;
         IF thisx - thisrad < xmin THEN xmin := thisx - thisrad;
         IF thisy + thisrad > ymax THEN ymax := thisy + thisrad;
         IF thisy - thisrad < ymin THEN ymin := thisy - thisrad;
         END;
      xrange := xmax - xmin;
      yrange := ymax - ymin;
      IF sizemode = yes
         THEN BEGIN
            IF ysize/xsize > yrange/xrange
               THEN factor := xsize/xrange
               ELSE factor := ysize/yrange;
            IF 2.0*radius*factor > maxdiam
               THEN factor := maxdiam/(2.0*radius);
            END
         ELSE factor := 1.0;
      CASE xoption OF
         atroot: BEGIN END;
         right: xtrans := xposition - xmin*factor;
         left: xtrans := xposition - xmax*factor;
         center: xtrans := xposition + (xrange/2.0 - xmax)*factor
         END;
      CASE yoption OF
         atroot: BEGIN END;
         above: ytrans := yposition - ymin*factor;
         below: ytrans := yposition - ymax*factor;
         center: ytrans := yposition + (yrange/2.0 - ymax)*factor
         END;
      FOR i := 1 TO nnodes DO BEGIN
         p := nodelist[i];
         p^.x := (p^.x*factor+xtrans)*xunitspercm;
         p^.y := (p^.y*factor+ytrans)*xunitspercm
         END;
      absradius := radius * factor * xunitspercm;
      xrange := xrange * factor * xunitspercm;
      yrange := yrange * factor * xunitspercm;
      FOR i := 1 TO 30 DO BEGIN
         angle := i * pi/15.0;
         xcircle[i] := SIN(angle) * absradius;
         ycircle[i] := -COS(angle) * absradius;
         END;
      END; (* hardwarecoordinates *)

   PROCEDURE loadfont(VAR font : fonttype);
      VAR i, charstart, hersheynum : INTEGER;
         ch : CHAR;
   BEGIN (* loadfont *)
      ch := 'A';
      i := 0;
      WHILE NOT(EOF(fontfile) OR (ch=' ')) DO BEGIN
         charstart := i+1;
         READ(fontfile, ch, ch, hersheynum,
            font[charstart+2], font[charstart+3]);
         font[charstart+1] := ORD(ch);
         i := charstart+3;
         REPEAT
            IF (i-(charstart+3)) MOD 10=0 THEN READLN(fontfile);
            i := i+1;
            READ(fontfile,font[i]);
            UNTIL ABS(font[i]) >= 10000;
         READLN(fontfile);
         font[charstart] := i+1;
         END;
      font[charstart] := 0;
      WRITELN(OUTPUT, '  FONT REQUIRES', i:5, ' INTEGERS OUT OF',
         fontsize:5, ' AVAILABLE');
      END; (* loadfont *)

   PROCEDURE labelsetup;
      VAR i : INTEGER;
   BEGIN (* labelsetup *)
      longestlablen := 0;
      FOR i := 1 TO nnodes DO
         IF nodelist[i]^.nch > longestlablen
            THEN longestlablen := nodelist[i]^.nch;
      IF longestlablen > 0
         THEN labelheight := 2.0*absradius/
                                   SQRT(SQR(longestlablen)+1.0)
         ELSE labelspresent := FALSE;
      END; (* labelsetup *)

   PROCEDURE plot(penstatus : penstatustype; xabs, yabs : REAL);
      VAR i, n, xrel, yrel, xinc, yinc, xhigh, yhigh,
            xlow, ylow, xlast, ylast : INTEGER;
         quadrant : CHAR;

      PROCEDURE pout(n : INTEGER);
      BEGIN (* pout *)
         IF n >= 1000
            THEN WRITE(plotfile, n:4)
            ELSE IF n >= 100
               THEN WRITE(plotfile, n:3)
               ELSE IF n >= 10
                  THEN WRITE(plotfile, n:2)
                  ELSE IF n >= 0
                     THEN WRITE(plotfile, n:1)
                     ELSE IF n >= -9
                        THEN WRITE(plotfile, n:2)
                        ELSE IF n >= -99
                           THEN WRITE(plotfile, n:3)
                           ELSE IF n >= -999
                              THEN WRITE(plotfile, n:4)
                              ELSE WRITE(plotfile, n:5);
         END; (* pout *)

      BEGIN (* plot *)
         yabs := yabs * aspectratio;
         CASE plotter OF
            calcomp: BEGIN
               IF penstatus = pendown
                  THEN WRITE(plotfile,'H')
                  ELSE WRITE(plotfile,'D');
               xrel := ROUND(xabs) - xnow;
               yrel := ROUND(yabs) - ynow;
               xnow := ROUND(xabs);
               ynow := ROUND(yabs);
               IF xrel > 0
                  THEN IF yrel > 0
                     THEN quadrant := 'P'
                     ELSE quadrant := 'T'
                  ELSE IF yrel > 0
                     THEN quadrant := 'X'
                     ELSE quadrant := '1';
               xrel := ABS(xrel);
               yrel := ABS(yrel);
               IF xrel > yrel
                  THEN n := (xrel DIV 255) + 1
                  ELSE n := (yrel DIV 255) + 1;
               xinc := xrel DIV n;
               yinc := yrel DIV n;
               xlast := xrel MOD n;
               ylast := yrel MOD n;
               xhigh := xinc DIV 16;
               yhigh := yinc DIV 16;
               xlow := xinc MOD 16;
               ylow := yinc MOD 16;
               FOR i := 1 TO n DO
                  WRITE(plotfile, quadrant, cchex[xhigh],
                     cchex[xlow], cchex[yhigh], cchex[ylow]);
               IF (xlast <> 0) OR (ylast <> 0)
                  THEN WRITE(plotfile, quadrant, cchex[0],
                     cchex[xlast], cchex[0], cchex[ylast]);
               END;
            tektronix: BEGIN
               IF penstatus = penup
                  THEN WRITE(plotfile, CHR(29));
               xnow := ROUND(xabs);
               ynow := ROUND(yabs);
               xhigh := xnow DIV 32;
               yhigh := ynow DIV 32;
               xlow := xnow MOD 32;
               ylow := ynow MOD 32;
               IF yhigh <> oldyhigh THEN WRITE(plotfile,
                                                 CHR(yhigh+32));
               IF (ylow <> oldylow) OR (xhigh <> oldxhigh)
                  THEN WRITE(plotfile, CHR(ylow+96));
               IF xhigh <> oldxhigh THEN WRITE(plotfile,
                                                 CHR(xhigh+32));
               WRITE(plotfile, CHR(xlow+64));
               oldxhigh := xhigh;
               oldxlow := xlow;
               oldyhigh := yhigh;
               oldylow := ylow;
               END;
            hp7470: BEGIN
               IF penstatus=pendown
                  THEN WRITE(plotfile, 'PD')
                  ELSE WRITE(plotfile, 'PU');
               pout(ROUND(xabs));
               WRITE(plotfile, ',');
               pout(ROUND(yabs));
               WRITELN(plotfile, ';');
               END;
            bauschlomb: BEGIN
               IF penstatus = pendown
                  THEN WRITE(plotfile, 'D ')
                  ELSE WRITE(plotfile, 'U ');
               pout(ROUND(xabs));
               WRITE(plotfile, ',');
               pout(ROUND(yabs));
               WRITELN(plotfile);
               END;
            other: BEGIN
               (* code for a pen move on a new plotter
                  goes here *)
               END
            END;
         END; (* plot *)

      PROCEDURE drawinternode(x1, y1, x2, y2, ndsize1,
                                           ndsize2 : REAL);
         VAR slope, xdiff, ydiff, xstart, ystart, xend,
                yend : REAL;

         FUNCTION trimend(thisend, farend, amount : REAL) : REAL;
         BEGIN (* trimend *)
            IF thisend > farend
               THEN trimend := thisend - amount
               ELSE trimend := thisend + amount;
            END; (* trimend *)

      BEGIN (* drawinternode *)
         IF x1 = x2
            THEN BEGIN
               xdiff := 0.0;
               ydiff := absradius;
               END
            ELSE IF y1 = y2
               THEN BEGIN
                  xdiff := absradius;
                  ydiff := 0.0;
                  END
               ELSE BEGIN
                  slope := (y2 - y1)/(x2 - x1);
                  xdiff := absradius/SQRT(SQR(slope) + 1.0);
                  ydiff := absradius/SQRT((1.0/SQR(slope)) + 1.0);
                  END;
         xstart := trimend(x1, x2, xdiff*ndsize1);
         ystart := trimend(y1, y2, ydiff*ndsize1);
         xend := trimend(x2, x1, xdiff*ndsize2);
         yend := trimend(y2, y1, ydiff*ndsize2);
         plot(penup, xstart, ystart);
         plot(pendown, xend, yend);
         END; (* drawinternode *)

      PROCEDURE drawnode(x, y, ndsize : REAL);
         VAR i : INTEGER;
      BEGIN (* drawnode *)
         plot(penup, x, y - absradius*ndsize);
         FOR i := 1 TO 30 DO plot(pendown, x+xcircle[i]*ndsize,
            y+ycircle[i]*ndsize);
         END; (* drawnode *)

      PROCEDURE plotlabel(p : nodeptr);
         TYPE pentype = (treepen, labelpen);
         VAR textlen, compress, xlab, ylab, x, y : REAL;

         PROCEDURE changepen(pen : pentype);
         BEGIN (* changepen *)
            CASE pen OF
               treepen: CASE plotter OF
                  hp7470: WRITELN(plotfile, 'SP1;');
                  bauschlomb: WRITELN(plotfile, 'P1');
                  other : BEGIN
                     (* code for a new plotter goes here *)
                     END
                  END;
               labelpen: CASE plotter OF
                  hp7470: WRITELN(plotfile, 'SP2;');
                  bauschlomb: WRITELN(plotfile, 'P2');
                  other : BEGIN
                     (* code for a new plotter goes here *)
                     END
                  END;
               END;
            END; (* changepen *)

         PROCEDURE plottext(pstring : plotstring; nchars : INTEGER;
            height, compress : REAL; VAR x, y : REAL;
            slope : REAL; VAR font : fonttype);
            CONST xstart = 10;
               ystart = 35;
            VAR i, j, code : INTEGER;
               sinslope, cosslope : REAL;

            PROCEDURE plotchar(VAR place : INTEGER);
               VAR coord : INTEGER;
                  heightfont, widthfont, xfactor, yfactor, xfont,
                  yfont, xplot, yplot : REAL;
                  penstatus : penstatustype;
            BEGIN (* plotchar *)
               heightfont := font[place+2];
               widthfont := font[place+3];
               yfactor := height/heightfont;
               xfactor := yfactor*compress;
               place := place+3;
               REPEAT
                  place := place+1;
                  coord := font[place];
                  IF coord > 0
                     THEN penstatus := pendown
                     ELSE penstatus := penup;
                  coord := ABS(coord);
                  coord := coord MOD 10000;
                  xfont := ((coord DIV 100) - xstart)*xfactor;
                  yfont := ((coord MOD 100) - ystart)*yfactor;
                  IF slope = 0.0
                     THEN BEGIN
                        xplot := x + xfont;
                        yplot := y + yfont;
                        END
                     ELSE BEGIN
                        xplot := x + xfont*cosslope
                                    + yfont*sinslope;
                        yplot := y - xfont*sinslope
                                    + yfont*cosslope;
                        END;
                  plot(penstatus, xplot, yplot);
                  UNTIL ABS(font[place]) >= 10000;
               x := xplot;
               y := yplot;
               END; (* plotchar *)

         BEGIN (* plottext *)
            IF slope <> 0.0 THEN BEGIN
               sinslope := SIN(slope);
               cosslope := COS(slope);
               END;
            FOR i := 1 TO nchars DO BEGIN
               code := ORD(pstring[i]);
               j := 1;
               WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                  j := font[j];
               IF font[j+1] = code THEN plotchar(j);
               END;
            END; (* plottext *)

         FUNCTION lengthtext(pstring : plotstring; nchars : INTEGER;
            height, compress : REAL; VAR font : fonttype) : REAL;
            VAR i, j, code, coord : INTEGER;
               sumlength, heightfont, widthfont, yfactor : REAL;
         BEGIN (* lengthtext *)
            sumlength := 0.0;
            FOR i := 1 TO nchars DO BEGIN
               code := ORD(pstring[i]);
               j := 1;
               WHILE (font[j+1] <> code) AND (font[j] <> 0) DO
                  j := font[j];
               IF font[j+1] = code THEN BEGIN
                  heightfont := font[j+2];
                  widthfont := font[j+3];
                  sumlength := sumlength + (height/heightfont)
                        *widthfont;
                  END;
               END;
               lengthtext := sumlength*compress;
            END; (* lengthtext *)

      BEGIN (* plotlabel *)
         IF penchange = yes THEN changepen(labelpen);
         compress := 1.0;
         textlen := lengthtext(p^.ndlabel, p^.nch,
            labelheight, compress, font);
         If textlen > p^.nch*labelheight THEN BEGIN
            compress := p^.nch*labelheight/textlen;
            textlen := p^.nch*labelheight;
            END;
         xlab := -textlen*labelsize/2.0;
         ylab := -labelheight*labelsize/2.0;
         x := p^.x + xlab*COS(rotation) + ylab*SIN(rotation);
         y := p^.y - xlab*SIN(rotation) + ylab*COS(rotation);
         plottext(p^.ndlabel, p^.nch, labelheight*labelsize,
                      compress, x, y, rotation, font);
         IF penchange = yes THEN changepen(treepen);
         END; (* plotlabel *)

      PROCEDURE drawtree(p : nodeptr);
      BEGIN (* drawtree *)
         IF p <> NIL THEN BEGIN
            IF p^.ancestor <> NIL THEN drawinternode(p^.ancestor^.x,
               p^.ancestor^.y, p^.x, p^.y, p^.ancestor^.nodesize,
               p^.nodesize);
            drawnode(p^.x, p^.y, p^.nodesize);
            IF labelspresent AND (p^.nch > 0) THEN plotlabel(p);
            drawtree(p^.descendant);
            drawtree(p^.sibling);
            END;
         END; (* drawtree *)

   PROCEDURE initplotter;
   BEGIN (* initplotter *)
      CASE plotter OF
         calcomp: BEGIN
            cchex[0] := 'C'; cchex[1] := 'D'; cchex[2] := 'H';
            cchex[3] := 'L'; cchex[4] := 'P'; cchex[5] := 'T';
            cchex[6] := 'X'; cchex[7] := '1'; cchex[8] := '5';
            cchex[9] := '9'; cchex[10] := '/'; cchex[11] := '=';
            cchex[12] := '#'; cchex[13] := '"'; cchex[14] := '''';
            cchex[15] := '@';
            xnow := 0;
            ynow := 0;
            WRITE(plotfile, 'CCCCCCCCCC');
            END;
         tektronix: BEGIN
            oldxhigh := -1;
            oldxlow := -1;
            oldyhigh := -1;
            oldylow := -1;
            WRITE(plotfile, CHR(27), CHR(12));
            END;
         hp7470: WRITELN(plotfile, 'IN;SP1;VS10.0;');
         other : BEGIN
            (* initialization code for a new plotter goes here *)
            END
         END;
      END; (* initplotter *)

      PROCEDURE finishplotter;
      BEGIN (* finishplotter *)
         CASE plotter OF
            calcomp: plot(penup, 0.0, yrange+50.0);
            hp7470: BEGIN
               plot(penup,0.0,0.0);
               WRITELN(plotfile,'SP;');
               END;
            tektronix: WRITELN(plotfile);
            other : BEGIN
               (* termination code for a new plotter goes here *)
               END
            END;
         END; (* finishplotter *)

   PROCEDURE garbagestack;
      VAR i : INTEGER;
   BEGIN (* garbagestack *)
      FOR i := 1 TO nnodes DO BEGIN
         nodelist[i]^.descendant := garbage;
         garbage := nodelist[i];
         END;
      END; (* garbagestack *)

BEGIN (* plotree *)
   WRITELN(OUTPUT, ' PLOTREE VERSION 1.1');
   WRITELN(OUTPUT, ' DISTRIBUTED WITH PHYLIP 2.7');
   WRITELN(OUTPUT, '  WRITTEN BY C. MEACHAM, BOTANY DEPT., ',
                                'UNIV. GEORGIA, ATHENS, GA 30602');
   WRITELN(OUTPUT, ' READING OF PARAMETERS BEGINS');
   RESET(parmfile);
   initparms;
   (* CLOSE(parmfile);   may have to change this *)
   WRITELN(OUTPUT, ' READING OF PARAMETERS COMPLETED');
   IF parmsok THEN BEGIN
      RESET(treefile);
      WHILE NOT EOF(treefile) DO BEGIN
         candraw := TRUE;
         WRITELN(OUTPUT, ' READING OF TREE BEGINS');
         readtree;
         WRITELN(OUTPUT, ' READING OF TREE COMPLETED');
         WRITELN(OUTPUT, ' CALCULATION OF TREE BEGINS');
         calctree(root);
         WRITELN(OUTPUT, ' CALCULATION OF TREE COMPLETED');
         IF candraw
            THEN BEGIN
               hardwarecoordinates;
               IF labelspresent THEN BEGIN
                  labelsetup;
                  IF NOT(fontloaded) AND labelspresent THEN BEGIN
                     WRITELN(OUTPUT, ' LOADING OF FONT BEGINS');
                     RESET(fontfile);
                     loadfont(font);
                     (* CLOSE(fontfile);   may have to change this *)
                     WRITELN(OUTPUT, ' LOADING OF FONT COMPLETED');
                     fontloaded := TRUE;
                     END;
                  END;
               WRITELN(OUTPUT, ' DRAWING OF TREE BEGINS');
               REWRITE(plotfile);
               initplotter;
               drawtree(root);
               finishplotter;
               WRITELN(OUTPUT, ' DRAWING OF TREE COMPLETED');
               END
            ELSE BEGIN
               WRITELN(OUTPUT, ' TREE CANNOT BE DRAWN WITH CURRENT ',
                         'PARAMETERS');
               WRITELN(OUTPUT, '  NODE DIAMETER MUST BE SMALLER OR ',
                         'INTERNODE LENGTH LARGER');
               END;
         garbagestack;
         END;
      END;
(*   CLOSE(treefile);   may have to change this *)
(*   CLOSE(plotfile);   may have to change this *)
   WRITELN(OUTPUT, ' EXECUTION TERMINATED');
   END. (* plotree *)
nt(font);
                     (* CLOSE(fontfile);   may have to change this *)
                     WRITELN(OUTPUT, ' LOADING